# copyright 2020-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import html
import secrets

import requests
from authlib.integrations.requests_client import OAuth2Session
from authlib.jose import jwt
from authlib.oauth2.rfc7636 import create_s256_code_challenge
from cubicweb import NoResultError
from cubicweb.pyramid.config import get_random_secret_key
from cubicweb.server.utils import crypt_password
from pyramid.authentication import IAuthenticationPolicy
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPFound,
    HTTPUnauthorized,
)
from pyramid.security import NO_PERMISSION_REQUIRED, remember
from pyramid.view import view_config
from zope.interface import implementer


@view_config(route_name="oauth2-start", permission=NO_PERMISSION_REQUIRED)
def start(request):
    rd = request.params.get("rd") or "/"
    return HTTPFound(request.registry["oauth2"].create_authorization_url(request, rd))


def find_user(cnx, config, token):
    user_login = token[config["oauth2-token-login"]]
    try:
        user = cnx.find("CWUser", login=user_login).one()
    except NoResultError:
        if not config["oauth2-register-user"]:
            raise HTTPUnauthorized("no such user in database")
        kwargs = {
            key: token[config[f"oauth2-token-{key}"]]
            for key in ("firstname", "surname", "email")
        }
        email = kwargs.pop("email")
        try:
            address = cnx.find("EmailAddress", address=email).one()
        except NoResultError:
            address = cnx.create_entity("EmailAddress", address=email)
        group = cnx.find("CWGroup", name=config["oauth2-default-group"]).one()
        user = cnx.create_entity(
            "CWUser",
            login=user_login,
            upassword=crypt_password(get_random_secret_key()),
            in_group=group,
            use_email=address,
            primary_email=address,
            **kwargs,
        )
        cnx.commit()
    return user.eid


@view_config(route_name="oauth2-callback", permission=NO_PERMISSION_REQUIRED)
def callback(request):
    try:
        state = request.params["state"]
    except KeyError:
        raise HTTPBadRequest("missing state")
    rd = request.params.get("rd", "/")
    oauth2_client = request.registry["oauth2"]
    token, rd = oauth2_client.fetch_token(request, state, rd)
    config = request.registry["cubicweb.config"]
    with request.registry["cubicweb.repository"].internal_cnx() as cnx:
        eid = oauth2_client.user_management(cnx, config, token)
    return HTTPFound(rd, headers=remember(request, eid))


class Oauth2:
    def __init__(
        self,
        client_id,
        client_secret,
        server_url=None,
        authorization_url=None,
        token_url=None,
        jwk_path=None,
        user_management=None,
    ):
        if not server_url and (not authorization_url or not token_url or not jwk_path):
            raise ValueError(
                "You should either set `url` or all "
                "`authorization_url`, `token_url` and `jwk_path`"
            )
        self._client_id = client_id
        self._client_secret = client_secret
        self._server_url = server_url
        self._authorization_url = authorization_url
        self._token_url = token_url
        self._session = self._create_session()
        if jwk_path:
            with open(jwk_path) as f:
                self._jwk = f.read()
        else:
            self._jwk = None
        if user_management is None:
            self.user_management = find_user

    def _create_session(self, state=None):
        return OAuth2Session(
            self._client_id,
            self._client_secret,
            scope="openid email profile",
            state=state,
        )

    def _read_openid_configuration(self):
        resp = requests.get(
            self._server_url.rstrip("/") + "/.well-known/openid-configuration",
        )
        resp.raise_for_status()
        config = resp.json()
        if not self._authorization_url:
            self._authorization_url = config["authorization_endpoint"]
        if not self._token_url:
            self._token_url = config["token_endpoint"]
        if not self._jwk:
            resp = requests.get(config["jwks_uri"])
            resp.raise_for_status()
            self._jwk = resp.json()

    @property
    def authorization_url(self):
        if not self._authorization_url:
            self._read_openid_configuration()
        return self._authorization_url

    @property
    def token_url(self):
        if not self._token_url:
            self._read_openid_configuration()
        return self._token_url

    @property
    def jwk(self):
        if not self._jwk:
            self._read_openid_configuration()
        return self._jwk

    def create_authorization_url(self, request, rd):
        redirect_uri = request.route_url("oauth2-callback", _query={"rd": rd})
        auth_url, state = self._session.create_authorization_url(
            self.authorization_url,
            redirect_uri=redirect_uri,
        )
        request.session["_oauth2_state"] = state
        return auth_url

    def fetch_token(self, request, state, rd):
        if state != request.session.get("_oauth2_state"):
            raise HTTPBadRequest("invalid state")
        del request.session["_oauth2_state"]
        redirect_uri = request.route_url("oauth2-callback", _query={"rd": rd})
        session = self._create_session(state)
        resp = session.fetch_token(
            self.token_url,
            authorization_response=request.url,
            redirect_uri=redirect_uri,
        )
        token = jwt.decode(resp["id_token"], self.jwk)
        return token, rd


class Oauth2PKCE(Oauth2):
    def create_authorization_url(self, request, rd):
        code_verifier = secrets.token_hex(24)
        code_challenge = create_s256_code_challenge(code_verifier)
        redirect_uri = request.route_url("oauth2-callback", _query={})

        auth_url, state = self._session.create_authorization_url(
            self.authorization_url,
            redirect_uri=redirect_uri,
            code_challenge=code_challenge,
            code_challenge_method="S256",
        )
        request.session["_oauth2_state"] = {
            "state": state,
            "rd": rd,
            "code_verifier": code_verifier,
        }
        return auth_url

    def fetch_token(self, request, state, rd):
        oauth2_state = request.session.get("_oauth2_state")
        if state != oauth2_state["state"]:
            raise HTTPBadRequest("invalid state")
        rd = oauth2_state["rd"]
        code_verifier = oauth2_state["code_verifier"]
        del request.session["_oauth2_state"]
        redirect_uri = request.route_url("oauth2-callback", _query={})
        session = self._create_session(state)
        resp = session.fetch_token(
            self.token_url,
            authorization_response=request.url,
            redirect_uri=redirect_uri,
            code_verifier=code_verifier,
        )
        token = jwt.decode(resp["id_token"], self.jwk)
        return token, rd


@implementer(IAuthenticationPolicy)
class Oauth2AuthenticationPolicy:
    def authenticated_userid(self, request):
        """
        When `oauth2-force-login` is set, force authentication by
        redirecting unauthenticated requests to /oauth2/start
        """
        if not request.registry["cubicweb.config"]["oauth2-force-login"]:
            return
        if request.authorization and request.authorization.authtype == "Cubicweb":
            # do not block signed requests from cwclientlib
            return
        if request.path in {
            request.route_path(r) for r in ("oauth2-callback", "oauth2-start")
        }:
            return
        raise HTTPUnauthorized(
            body=(
                "<!DOCTYPE html><html><head>"
                '<meta http-equiv="refresh" content="0; url={}" />'
                "</head></html>"
            ).format(
                html.escape(
                    request.route_url(
                        "oauth2-start",
                        _query={"rd": request.url},
                    )
                )
            )
        )

    def effective_principals(self, request):
        return ()

    def remember(self, request, principal, **kw):
        return ()

    def forget(self, request):
        return ()


def includeme(config):
    cwconfig = config.registry["cubicweb.config"]
    Oauth2Class = Oauth2
    if cwconfig["oauth2-pkce-workflow"]:
        Oauth2Class = Oauth2PKCE
    config.registry["oauth2"] = Oauth2Class(
        cwconfig["oauth2-client-id"],
        cwconfig["oauth2-client-secret"],
        server_url=cwconfig["oauth2-server-url"],
        authorization_url=cwconfig["oauth2-authorization-url"],
        token_url=cwconfig["oauth2-token-url"],
        jwk_path=cwconfig["oauth2-jwk-path"],
    )
    policy = Oauth2AuthenticationPolicy()
    config.registry["cubicweb.authpolicy"]._policies.append(policy)
    config.add_route("oauth2-start", "/oauth2/start")
    config.add_route("oauth2-callback", "/oauth2/callback")
    config.scan(__name__)



# copyright 2020-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import urllib.parse
from unittest.mock import MagicMock

import responses
from authlib.jose import jwt
from cubicweb.devtools import BASE_URL
from cubicweb_web.devtools.testlib import PyramidWebCWTC

from cubicweb_oauth2.auth import Oauth2, Oauth2PKCE

HERE = os.path.dirname(__file__)


def create_token(payload):
    # use a RSA 256 keypair for tests, generated with:
    # ssh-keygen -t rsa -b 4096 -m PEM -f jwt.key
    # openssl rsa -in jwt.key -pubout -outform PEM -out jwt.pub
    with open(os.path.join(HERE, "jwt.key")) as f:
        privkey = f.read()
    return jwt.encode({"alg": "RS256"}, payload, privkey).decode()


class AuthenticationTC(PyramidWebCWTC):
    settings = {
        **PyramidWebCWTC.settings,
        "cubicweb.includes": ["cubicweb.pyramid.auth", "cubicweb.pyramid.session"],
    }

    @classmethod
    def init_config(cls, config):
        super().init_config(config)
        config.global_set_option("oauth2-enabled", True)
        config.global_set_option("oauth2-client-id", "id")
        config.global_set_option("oauth2-client-secret", "secret")
        config.global_set_option("oauth2-authorization-url", "https://provider/auth")
        config.global_set_option("oauth2-token-url", "https://provider/token")
        config.global_set_option("oauth2-jwk-path", os.path.join(HERE, "jwt.pub"))
        config.global_set_option("oauth2-provider-name", "Logilab")

    def test_login_page(self):
        resp = self.webapp.get("/login?postlogin_path=/schema", status=200)
        assert (
            f'<a href="{BASE_URL}oauth2/start?rd=%2Fschema">'
            f'<button class="validateButton" type="button" '
            f'value="Log in with Logilab">Log in with Logilab</button></a>'
        ).encode() in resp.body

    def test_login_page_auto_login(self):
        self.set_option("oauth2-auto-login", True)
        try:
            resp = self.webapp.get("/login?postlogin_path=/schema", status=302)
        finally:
            self.set_option("oauth2-auto-login", False)
        assert resp.location == f"{BASE_URL}oauth2/start?rd=%2Fschema"

    def test_force_login(self):
        self.set_option("oauth2-force-login", True)
        try:
            resp = self.webapp.get("/poulet", status=401)
        finally:
            self.set_option("oauth2-force-login", False)
        assert (
            resp.body
            == (
                f'<!DOCTYPE html><html><head><meta http-equiv="refresh" content="0; '
                f'url={BASE_URL}oauth2/start?rd={urllib.parse.quote(BASE_URL, safe="")}poulet" />'
                f"</head></html>"
            ).encode()
        )

    def test_force_login_signedrequest(self):
        # we should not block requests with signedrequest token authentication
        self.set_option("oauth2-force-login", True)
        try:
            self.webapp.get(
                "/poulet", headers={"Authorization": "Cubicweb foo"}, status=404
            )
        finally:
            self.set_option("oauth2-force-login", False)

    @responses.activate
    def test_full_login(self):
        resp = self.webapp.get("/oauth2/start?rd=/page", status=302)
        url = urllib.parse.urlparse(resp.location)
        assert url.scheme + "://" + url.netloc + url.path == "https://provider/auth"
        qs = urllib.parse.parse_qs(url.query)
        state = qs["state"][0]
        assert qs == {
            "response_type": ["code"],
            "client_id": ["id"],
            "redirect_uri": [f"{BASE_URL}oauth2/callback?rd=%2Fpage"],
            "scope": ["openid email profile"],
            "state": [state],
        }
        resp = self.webapp.get(
            "/oauth2/callback?rd=%2Fpage&state=invalid",
            status=400,
        )
        assert b"invalid state" in resp.body
        # this is a typical simplified keycloak token
        token = {
            "aud": "test",
            "email": "jdoe@logilab.fr",
            "family_name": "Doe",
            "given_name": "John",
            "name": "John Doe",
            "preferred_username": "jdoe",
            "sub": "6e349788-4b4a-4176-bc0a-81b1c48a675e",
        }
        responses.add(
            responses.POST,
            "https://provider/token",
            json={"id_token": create_token(token)},
            status=200,
        )
        resp = self.webapp.get(
            f"/oauth2/callback?rd=%2Fpage&state={state}&code=sesame",
            status=302,
        )
        assert resp.location == f"{BASE_URL}page"
        with self.admin_access.cnx() as cnx:
            user = cnx.find("CWUser", login="jdoe").one()
            assert (user.surname, user.firstname) == ("Doe", "John")
            assert [a.address for a in user.use_email] == ["jdoe@logilab.fr"]


def register_provide_response_ok(responses):
    responses.add(
        responses.GET,
        "https://provider/.well-known/openid-configuration",
        json={
            "authorization_endpoint": "https://provider/auth",
            "token_endpoint": "https://provider/token",
            "jwks_uri": "https://provider/certs",
        },
        status=200,
    )
    dummy_jwk = {"keys": [{"alg": "RS256", "kty": "RSA", "n": "dummy"}]}
    responses.add(
        responses.GET,
        "https://provider/certs",
        json=dummy_jwk,
        status=200,
    )
    return dummy_jwk


def parse_url(url):
    url_parser = urllib.parse.urlparse(url)
    return urllib.parse.parse_qs(url_parser.query)


class Oauth2TC(unittest.TestCase):
    def test_required_params(self):
        msg = (
            "You should either set `url` or all "
            "`authorization_url`, `token_url` and `jwk_path`"
        )
        with self.assertRaises(ValueError) as cm:
            Oauth2("id", "secret")
        assert str(cm.exception) == msg

        with self.assertRaises(ValueError) as cm:
            Oauth2("id", "secret", authorization_url="/auth")
        assert str(cm.exception) == msg

        # This should not raise
        Oauth2("id", "secret", server_url="/")
        Oauth2(
            "id",
            "secret",
            token_url="/token",
            authorization_url="/auth",
            jwk_path=os.path.join(HERE, "jwt.pub"),
        )

    @responses.activate
    def test_url_discovery(self):
        dummy_jwk = register_provide_response_ok(responses)
        client_id = "id"
        oauth2 = Oauth2(client_id, "secret", server_url="https://provider")
        assert oauth2.authorization_url == "https://provider/auth"
        assert oauth2.token_url == "https://provider/token"
        assert oauth2.jwk == dummy_jwk
        request = MagicMock()
        rd = "/42"
        authorization_url = oauth2.create_authorization_url(request, rd)
        params = parse_url(authorization_url)
        expected_params = {
            "response_type",
            "client_id",
            "redirect_uri",
            "scope",
            "state",
        }

        assert set(params.keys()) == expected_params
        assert params["client_id"] == [client_id]
        assert params["scope"] == ["openid email profile"]


class Oauth2PKCETC(unittest.TestCase):
    @responses.activate
    def test_create_authorization_url(self):
        register_provide_response_ok(responses)
        oauth2 = Oauth2PKCE("id", "secret", server_url="https://provider")
        request = MagicMock()
        rd = "/42"
        authorization_url = oauth2.create_authorization_url(request, rd)
        params = parse_url(authorization_url)
        expected_params = {
            "response_type",
            "client_id",
            "redirect_uri",
            "scope",
            "state",
            "code_challenge",
            "code_challenge_method",
        }
        assert set(params.keys()) == expected_params
        assert params["code_challenge_method"] == ["S256"]
        assert request.session.__setitem__.called
        state = request.session.__setitem__.call_args_list[0].args[1]
        assert params["state"][0] == state["state"]



# copyright 2020-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-oauth2 views/forms/actions/components for web ui"""

from cubicweb import _, tags
from cubicweb_web.formwidgets import Button
from cubicweb_web.views.basetemplates import LogForm, LogFormView
from pyramid.httpexceptions import HTTPFound


class LoginWithOauth2Button(Button):
    def __init__(self, label, url):
        super().__init__(label=label)
        self._url = url

    def render(self, *args, **kwargs):
        button = super().render(*args, **kwargs)
        return tags.div(tags.a(button, href=self._url, escapecontent=False))


def start_login_url(cw):
    return cw._request.route_url(
        "oauth2-start",
        _query={"rd": cw._request.params.get("postlogin_path")},
    )


def login_button(cw):
    name = cw.vreg.config["oauth2-provider-name"]
    return LoginWithOauth2Button(
        _("Log in with %s") % (name,),
        start_login_url(cw),
    )


class Oauth2LogFormView(LogFormView):
    def call(self, *args, **kwargs):
        config = self._cw.vreg.config
        if config["oauth2-enabled"]:
            if config["oauth2-auto-login"]:
                raise HTTPFound(start_login_url(self._cw))

            cw = self._cw
            form = cw.vreg["forms"].select("logform", cw)
            if not any(isinstance(b, LoginWithOauth2Button) for b in form.form_buttons):
                form.form_buttons.append(login_button(self._cw))
        return super().call(*args, **kwargs)


class Oauth2LogForm(LogForm):
    @property
    def form_buttons(self):
        buttons = super().form_buttons
        if self._cw.vreg.config["oauth2-enabled"]:
            buttons.append(login_button(self._cw))
        return buttons


def registration_callback(vreg):
    vreg.register_and_replace(Oauth2LogFormView, LogFormView)
    if "bootstrap" not in vreg.config.cubes():
        vreg.register_and_replace(Oauth2LogForm, LogForm)



"""cubicweb-oauth2 application package

Oauth2/OpenID authentication for cubicweb
"""


def includeme(config):
    cwconfig = config.registry["cubicweb.config"]
    if not cwconfig["oauth2-enabled"]:
        return
    config.include(".auth")



# copyright 2020-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

options = (
    (
        "oauth2-enabled",
        {
            "type": "yn",
            "default": False,
            "help": "Enable Oauth2 client",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-pkce-workflow",
        {
            "type": "yn",
            "default": False,
            "help": "Use PKCE authorization workflow",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-auto-login",
        {
            "type": "yn",
            "default": False,
            "help": "Skip the login page and try to login directly",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-force-login",
        {
            "type": "yn",
            "default": False,
            "help": "Force login on all unauthenticated requests",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-provider-name",
        {
            "type": "string",
            "default": "Oauth2",
            "help": "Name of the Oauth2 provider displayed in login page",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-server-url",
        {
            "type": "string",
            "default": "",
            "help": (
                "Base URL for Oauth2 server. Will look at "
                "/.well-known/openid-configuration for others urls."
            ),
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-authorization-url",
        {
            "type": "string",
            "default": "",
            "help": (
                "Oauth2 authorization url. "
                "Set this to avoid a request on oauth2-server-url at startup."
            ),
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-token-url",
        {
            "type": "string",
            "default": "",
            "help": (
                "Oauth2 token url "
                "Set this to avoid a request on oauth2-server-url at startup."
            ),
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-jwk-path",
        {
            "type": "string",
            "default": "",
            "help": (
                "Oauth2 server public key path. "
                "Set this to avoid a request to oauth2-server-url at startup."
            ),
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-client-id",
        {
            "type": "string",
            "default": "",
            "help": "Oauth2 client id",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-client-secret",
        {
            "type": "string",
            "default": "",
            "help": "Oauth2 client secret",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-register-user",
        {
            "type": "yn",
            "default": True,
            "help": "Automatically register if it does not exists yet",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-token-login",
        {
            "type": "string",
            "default": "preferred_username",
            "help": "Field on JWT token corresponding to CubicWeb user login",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-default-group",
        {
            "type": "string",
            "default": "guests",
            "help": "Default CWGroup when registering a new user",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-token-firstname",
        {
            "type": "string",
            "default": "given_name",
            "help": "Field on JWT token corresponding to CubicWeb user firstname",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-token-surname",
        {
            "type": "string",
            "default": "family_name",
            "help": "Field on JWT token corresponding to CubicWeb user surname",
            "group": "oauth2",
            "level": 5,
        },
    ),
    (
        "oauth2-token-email",
        {
            "type": "string",
            "default": "email",
            "help": "Field on JWT token corresponding to CubicWeb email address",
            "group": "oauth2",
            "level": 5,
        },
    ),
)



"""cubicweb-oauth2 application packaging information"""

from importlib.metadata import metadata

modname = "cubicweb_oauth2"
distname = "cubicweb-oauth2"

cubicweb_metadata = metadata(distname).json

version = cubicweb_metadata["version"]
numversion = [int(number) for number in version.split(".")]

description = cubicweb_metadata["description"]
author, author_email = cubicweb_metadata["author_email"].split()
web = cubicweb_metadata["project_url"][-1].split(", ")[-1]
license = cubicweb_metadata["license"]
classifiers = cubicweb_metadata["classifier"]
